{
@abstract(provides HTML doc generator object)
@author(Erwin Scheuch-Heilig (ScheuchHeilig@t-online.de))
@author(Marco Schmidt (marcoschmidt@geocities.com))
@created(1 Sep 1998)
@lastmod(10 Apr 2000)

Implements an object to generate HTML documentation, overriding many of
@link(TDocGenerator)'s virtual methods.
}

{$V-}

unit html;

interface

uses
  appinfo, arrays, filestre, gendoc, items, numbers, objects, parsing,
  streams, texts, time;

type
  THTMLColorIndex = (HTML_BACKGROUND, HTML_TEXT, HTML_LINK, HTML_VLINK, HTML_ALINK,
    HTML_TABLE_BACKGROUND, HTML_TABLE_TEXT);

type
  { pointer to @link(THTMLDocGenerator) }
  PHTMLDocGenerator = ^THTMLDocGenerator;
  { @abstract(generates HTML documentation)
    Extends @link(TDocGenerator) and overwrites many of its methods to generate
    output in HTML (HyperText Markup Language) format.
    This type of output is well suited to be read with a web browser at the
    computer, as a reference manual that does not have to be printed.
    For printed output, use @link(Tex.TTexDocGenerator). }
  THTMLDocGenerator = object(TDocGenerator)
    { Converts C to HTML, simply copying normal characters and converting
      special characters like the ampersand to escape sequences. }
    function ConvertChar(C: Char): String; virtual;
    { Returns a link to an anchor within a document. HTML simply concatenates
      the strings with a "#" character between them. }
    function CreateLink(Item: PItem): String; virtual;
    { Creates a valid HTML link, starting with an anchor that points to Link,
      encapsulating the text ItemName in it. }
    function CreateReferencedLink(ItemName, Link: String): String; virtual;
    { Returns HTML file extension ".htm". }
    function GetFileExtension: String; virtual;
    { }
{        function GetLink(ItemReference, ItemName: String): String; virtual;}
    { Writes information on doc generator to current output stream,
      including link to pasdoc homepage. }
    procedure WriteAppInfo;
    { Writes authors to output, at heading level HL. Will not write anything
      if collection of authors is not assigned or empty. }
    procedure WriteAuthors(HL: Byte; Authors: PStringCollection);
    { Writes a single class, interface or object CIO to output, at heading
      level HL. }
    procedure WriteCIO(HL: Byte; var CIO: TCIO); virtual;
    { Calls @link(WriteCIO) with each element in the argument collection C,
      using heading level HL. }
    procedure WriteCIOs(HL: Byte; C: PItemCollection); virtual;
    procedure WriteCIOSummary(HL: Byte; C: PItemCollection); virtual;
    procedure WriteCodeWithLinks(P: PItem; S: String);
    { Writes a comment S to output file. }
    procedure WriteComment(const S: String); virtual;
    { Writes dates Created and LastMod at heading level HL to output
      (if at least one the two has a value assigned). }
    procedure WriteDates(HL: Byte; Created, LastMod: PString); virtual;
    { The method that does everything - writes documentation for all units
      and creates overview files. }
    procedure WriteDocumentation; virtual;
    { Writes the beginning of the HTML document, including opening HTML element,
      a complete HEAD element and an opening BODY element.
      See @link(WriteEndOfDocument). }
    procedure WriteDocumentHeadline; virtual;
    { Writes an empty table cell, '&nbsp;'. }
    procedure WriteEmptyCell;
    { Writes the end of an HTML anchor, '</a>'. }
    procedure WriteEndOfAnchor;
    { See @link(WriteDocumentHeadline). }
    procedure WriteEndOfDocument;
    procedure WriteEndOfLink;
    { Finishes an HTML paragraph element by writing a closing P tag. }
    procedure WriteEndOfParagraph;
    { Finishes an HTML table cell by writing a closing TD tag. }
    procedure WriteEndOfTableCell;
    { Finishes an HTML table by writing a closing TABLE tag. }
    procedure WriteEndOfTable;
    { Finishes an HTML table row by writing a closing TR tag. }
    procedure WriteEndOfTableRow;
    { Writes information on functions and procedures or methods of a unit or
      class, interface or object to output.
      If argument Methods is true, they will be considered methods of a class,
      interface or object, otherwise they're considered functions or procedures
      of a unit.
      The functions are stored in the FuncsProcs argument. }
    procedure WriteFuncsProcs(HL: Byte; Methods: Boolean;
      FuncsProcs: PMethodCollection); virtual;
    { Writes heading S to output, at heading level I.
      For HTML, only levels 1 to 6 are valid, so that values smaller
      than 1 will be set to 1 and arguments larger than 6 are set to 6.
      The string S will then be enclosed in an element from H1 to H6,
      according to the level. }
    procedure WriteHeading(I: Byte; const S: String); virtual;
    procedure WriteItemDescription(Item: PItem);
    procedure WriteItems(Order: Byte; Heading: String; I: PItemCollection); virtual;
    procedure WriteOverviewFiles;
    procedure WriteParagraph(HL: Byte; S: String; T: PText);
    procedure WriteProperties(HL: Byte; P: PPropertyCollection); virtual;
    { Writes an opening A element, including a name attribute given by the
      argument. }
    procedure WriteStartOfAnchor(const Name: String);
    procedure WriteStartOfDocument(Name: String); virtual;
    procedure WriteStartOfLink(const Name: String);
    { Starts an HTML paragraph element by writing an opening P tag. }
    procedure WriteStartOfParagraph;
    procedure WriteStartOfTableCell;
    procedure WriteStartOfTable1Column(T: String);
    procedure WriteStartOfTable2Columns(T1, T2: String);
    procedure WriteStartOfTable3Columns(T1, T2, T3: String);
    procedure WriteStartOfTableRow;
    procedure WriteUnit(HL: Byte; U: PUnit); virtual;
    procedure WriteUnitDescription(HL: Byte; D: PText); virtual;
    { Creates an output stream that lists up all units and short descriptions. }
    procedure WriteUnitOverviewFile;
  end;

const
  { background color of a table header row; a light gray slightly darker
    than the light gray of @link(HTML_HEADER_BACKGROUND_COLOR)  }
  HTML_ROW_BACKGROUND_COLOR: String[6] = 'efefef';
  { background color of a normal table row; a light gray slightly lighter
    than the light gray of @link(HTML_ROW_BACKGROUND_COLOR) }
  HTML_HEADER_BACKGROUND_COLOR: String[6] = 'e0e0e0';

implementation

uses
  msg;
{
html things to be customized
- standard background color (white)
- background color for table headings and overview list at the top of each file (light gray)
- background color for normal table cells (light gray, slightly lighter than the above)
- standard foreground color (black)
- unused link color (blue)
- used link color (purple)
- link color while being clicked on (red)
- normal font (Times Roman)
- heading font (Helvetica)
- code font (Courier)
}

{$I platform.inc}

function THTMLDocGenerator.ConvertChar(C: Char): String;
begin
  case C of
    { characters already in use by HTML }
    '<': ConvertChar := '&lt;'; { less than }
    '>': ConvertChar := '&gt;'; { greater than }
    '&': ConvertChar := '&amp;'; { ampersand }
    '"': ConvertChar := '&quot;'; { quote }
    { some German special characters }
    #$e4: ConvertChar := '&auml;'; {  (small Umlaut a) }
    #$f6: ConvertChar := '&ouml;'; {  (small Umlaut o) }
    #$fc: ConvertChar := '&uuml;'; {  (small Umlaut u) }
    #$c4: ConvertChar := '&Auml;'; {  (large Umlaut A) }
    #$d6: ConvertChar := '&Ouml;'; {  (large Umlaut O) }
    #$dc: ConvertChar := '&Uuml;'; { U (large Umlaut U) }
    #$df: ConvertChar := '&szlig;'; {  (sz) }
    { more special characters to be added here }
    { otherwise, simply return the argument character }
    else ConvertChar := C;
  end;
end;

function THTMLDocGenerator.CreateLink(Item: PItem): String;
begin
  CreateLink := '';
  if (not Assigned(Item))
  then exit;
  if Assigned(Item^.MyUnit)
  then
    begin
      if Assigned(Item^.MyObject)
      then
        begin
          { it's a method, a field or a property - only those
            have MyObject initialized }
          CreateLink := Item^.MyObject^.Name + GetFileExtension + '#' +
            AnchorToString(Item^.AnchorNumber);
        end
      else
        begin
          if (TypeOf(TCIO) = TypeOf(Item^))
          then
            begin
              { it's an object / a class }
              CreateLink := Item^.Name + GetFileExtension;
            end
          else
            begin
              { it's a constant, a variable, a type or a function / procedure }
              CreateLink := Item^.MyUnit^.FullLink + '#' +
                AnchorToString(Item^.AnchorNumber);
            end;
        end;
    end
  else
    begin
      { it's a unit - only units don't have a MyUnit pointer }
      CreateLink := Item^.Name + GetFileExtension;
    end;
end;

function THTMLDocGenerator.CreateReferencedLink(ItemName, Link: String): String;
begin
  CreateReferencedLink := '<a href="' + Link + '">' + ItemName + '</a>';
end;

function THTMLDocGenerator.GetFileExtension: String;
begin
  { '.html' makes DOS version unhappy - welcome to the past ;-) }
  GetFileExtension :=
    {$ifdef OS_DOS}
    '.htm';
    {$else}
    '.html';
    {$endif}
end;

{
function THTMLDocGenerator.GetLink(ItemReference, ItemName: String): String;
begin
  GetLink := '<a href="' + ItemReference + '">' + ItemName + '</a>';
end;
}

{
 A mit Accent grave &Agrave; &#192; 
 A mit Accent acute &Aacute; &#193; 
 A mit Circumflex &Acirc; &#194; 
 A mit Tilde &Atilde; &#195; 
 A Umlaut &Auml; &#196; 
 A mit Ring &Aring; &#197; 
 A mit legiertem E &AElig; &#198; 
 C mit Hkchen &Ccedil; &#199; 
 E mit Accent grave &Egrave; &#200; 
 E mit Accent acute &Eacute; &#201; 
 E mit Circumflex &Ecirc; &#202; 
 E Umlaut &Euml; &#203; 
 I mit Accent grave &Igrave; &#204; 
 I mit Accent acute &Iacute; &#205; 
 I mit Circumflex &Icirc; &#206; 
 I Umlaut &Iuml; &#207; 
 Eth (islndisch) &ETH; &#208; 
 N mit Tilde &Ntilde; &#209; 
 O mit Accent grave &Ograve; &#210; 
 O mit Accent acute &Oacute; &#211; 
 O mit Circumflex &Ocirc; &#212; 
 O mit Tilde &Otilde; &#213; 
 O Umlaut &Ouml; &#214; 
 Mal-Zeichen &times; &#215; 
 O mit Schrgstrich &Oslash; &#216; 
 U mit Accent grave &Ugrave; &#217; 
 U mit Accent acute &Uacute; &#218; 
 U mit Circumflex &Ucirc; &#219; 
 U Umlaut &Uuml; &#220; 
 Y mit Accent acute &Yacute; &#221; 
 THORN (islndisch) &THORN; &#222; 
 scharfes S &szlig; &#223; 
 a mit Accent grave &agrave; &#224; 
 a mit Accent acute &aacute; &#225; 
 a mit Circumflex &acirc; &#226; 
 a mit Tilde &atilde; &#227; 
 a Umlaut &auml; &#228; 
 a mit Ring &aring; &#229; 
 a mit legiertem e &aelig; &#230; 
 c mit Hkchen &ccedil; &#231; 
 e mit Accent grave &egrave; &#232; 
 e mit Accent acute &eacute; &#233; 
 e mit Circumflex &ecirc; &#234; 
 e Umlaut &euml; &#235; 
 i mit Accent grave &igrave; &#236; 
 i mit Accent acute &iacute; &#237; 
 i mit Circumflex &icirc; &#238; 
 i Umlaut &iuml; &#239; 
 eth (islndisch) &eth; &#240; 
 n mit Tilde &ntilde; &#241; 
 o mit Accent grave &ograve; &#242; 
 o mit Accent acute &oacute; &#243; 
 o mit Circumflex &ocirc; &#244; 
 o mit Tilde &otilde; &#245; 
 o Umlaut &ouml; &#246; 
 Divisions-Zeichen &divide; &#247; 
 o mit Schrgstrich &oslash; &#248; 
 u mit Accent grave &ugrave; &#249; 
 u mit Accent acute &uacute; &#250; 
 u mit Circumflex &ucirc; &#251; 
 u Umlaut &uuml; &#252; 
 y mit Accent acute &yacute; &#253; 
 thorn (islndisch) &thorn; &#254; 
 y Umlaut &yuml; &#255; 
}

procedure THTMLDocGenerator.WriteAppInfo;
var
  T: TTime;
begin
  { check if user does not want a link to the pasdoc homepage }
  if NoGeneratorInfo
  then exit;
  { get current date/time }
  T.Init;
  { write a horizontal line, pasdoc version and a link to the pasdoc
    homepage }
  Stream^.WriteLine('<hr><em>Generated by <a href="' + AppInfo.APP_HOMEPAGE +
    '">' + AppInfo.APP_NAME_AND_VERSION + '</a> on ' +
    T.GetDefaultString + '</em>');
end;

procedure THTMLDocGenerator.WriteAuthors(HL: Byte; Authors: PStringCollection);
var
  I: LongInt;
  P: PString;
  S1: String;
  S2: String;
  EmailAddress: String;
begin
  if (not Assigned(Authors)) or (Authors^.Count < 1)
  then exit;
  if (Authors^.Count = 1)
  then WriteHeading(HL, GetString(TEXT_AUTHOR))
  else WriteHeading(HL, GetString(TEXT_AUTHORS));
  for I := 0 to Authors^.Count - 1 do
  begin
    P := Authors^.At(I);
    WriteStartOfParagraph;
    if Assigned(P) then
    begin
      if ExtractEmailAddress(P^, S1, S2, EmailAddress) then
      begin
        WriteString(S1);
        Stream^.WriteString('<a href="mailto:' + EmailAddress + '"><code>');
        WriteString(EmailAddress);
        Stream^.WriteString('</code></a>');
        WriteString(S2);
      end
      else
      begin
      	WriteString(P^);
      end;
    end;
    WriteEndOfParagraph;
  end;
end;

procedure THTMLDocGenerator.WriteCIO(HL: Byte; var CIO: TCIO);
var
  I: LongInt;
  P: PString;
  S: String;
begin
  WriteDocumentHeadline;
  case CIO.MyType of
    CIO_CLASS: S := GetString(TEXT_CLASS);
    CIO_INTERFACE: S := GetString(TEXT_INTERFACE);
    CIO_OBJECT: S := GetString(TEXT_OBJECT);
    else S := '';
  end;
  S := S + ' ' + CIO.Name;
  WriteStartOfAnchor(AnchorToString(CIO.AnchorNumber));
  WriteHeading(HL, S);
  WriteEndOfAnchor;
  { write unit link }
  if Assigned(CIO.MyUnit) then
  begin
    WriteHeading(HL + 1, GetString(TEXT_UNIT));
    Stream^.WriteString('<a href="' + CIO.MyUnit^.FullLink + '">' +
      CIO.MyUnit^.Name + '</a><br>');
  end;
  { write declaration link }
  WriteHeading(HL + 1, GetString(TEXT_DECLARATION));
  Stream^.WriteString('<code>type ' + CIO.Name + ' = ');
  case CIO.MyType of
    CIO_CLASS: Stream^.WriteString('class');
    CIO_INTERFACE: Stream^.WriteString('interface');
    else Stream^.WriteString('object');
  end;
  if Assigned(CIO.Ancestors) and (CIO.Ancestors^.Count > 0) then
  begin
    Stream^.WriteString('(');
    for I := 0 to CIO.Ancestors^.Count - 1 do
    begin
      P := PString(CIO.Ancestors^.At(I));
      S := SearchLink(P^, nil);
      Stream^.WriteString(S);
      if (I <> CIO.Ancestors^.Count - 1)
      then Stream^.WriteString(', ');
    end;
    Stream^.WriteString(')');
  end;
  Stream^.WriteString('</code><br>');
  WriteDescription(HL + 1, GetString(TEXT_DESCRIPTION), CIO);
  WriteItems(HL + 1, GetString(TEXT_FIELDS), CIO.Fields);
  WriteFuncsProcs(HL + 1, true, CIO.Methods);
  if (CIO.MyType <> CIO_OBJECT)
  then WriteProperties(HL + 1, CIO.Properties);
  WriteAuthors(HL + 1, CIO.Authors);
  WriteDates(HL + 1, CIO.Created, CIO.LastMod);
  WriteAppInfo;
  WriteEndOfDocument;
end;

procedure THTMLDocGenerator.WriteCIOs(HL: Byte; C: PItemCollection);
var
  I: LongInt;
  P: PCIO;
begin
  if (not Assigned(C)) or (C^.Count < 1)
  then exit;
  for I := 0 to C^.Count - 1 do
  begin
    P := C^.At(I);
    if (not CreateStream(P^.Name))
    then
      begin
        PrintLn(1, 'Error: Could not create class/interface/object doc file.');
        Continue;
      end;
    PrintLn(3, 'Creating class/interface/object file "' + Stream^.GetName + '"...');
    WriteCIO(HL, P^);
  end;
  CloseStream;
end;

procedure THTMLDocGenerator.WriteCIOSummary(HL: Byte; C: PItemCollection);
var
  I: LongInt;
  P: PCIO;
begin
  WriteHeading(HL, GetString(TEXT_CIO));
  if (not Assigned(C)) or (C^.Count < 1)
  then
    begin
      WriteStartOfParagraph;
      WriteString(GetString(TEXT_NONE));
      WriteEndOfParagraph;
      exit;
    end;
  WriteStartOfTable2Columns(GetString(TEXT_NAME), GetString(TEXT_DESCRIPTION));
  for I := 0 to C^.Count - 1 do
  begin
    P := C^.At(I);
    WriteStartOfTableRow;
    { name of class/interface/object and unit }
    WriteStartOfTableCell;
    WriteString(GetCIOTypeName(P^.MyType) + ' ');
    WriteStartOfLink(P^.FullLink);
    WriteString(P^.Name);
    WriteEndOfLink;
    WriteEndOfTableCell;
    { description of class/interface/object }
    WriteStartOfTableCell;
    if Assigned(P^.Description) and (P^.Description^.Content > 0)
    then WriteText(P^.Description)
    else Stream^.WriteString('&nbsp;');
    WriteEndOfTableCell;
    WriteEndOfTableRow;
  end;
  WriteEndOfTable;
end;

procedure THTMLDocGenerator.WriteCodeWithLinks(P: PItem; S: String);
var
  FoundColon: Boolean;
  FoundItem: PItem;
  I: LongInt;
  L: String;
  N: Integer;
  S1: String;
  S2: String;
  S3: String;
begin
  Stream^.WriteString('<code>');
  I := 1;
  FoundColon := false;
  while (I <= Length(S)) do
  begin
    if (S[I] in ['_', 'A'..'Z', 'a'..'z'])
    then
      begin
        { assemble item }
        L := '';
        repeat
          L :=  L + S[I];
          Inc(I);
        until (I > Length(S)) or
          (not (S[I] in ['.', '_', '0'..'9', 'A'..'Z', 'a'..'z']));
        { search for item of name  L }
        if FoundColon and (SplitLink(L, S1, S2, S3, N))
        then
          begin
            FoundItem := P^.FindName(S1, S2, S3, N);
            if (not Assigned(FoundItem))
            then
              begin
                FoundItem := FindGlobal(S1, S2, S3, N);
              end;
            if (FoundItem = P)
            then FoundItem := nil;
          end
        else FoundItem := nil;
        if (Assigned(FoundItem))
        then
          begin
            WriteStartOfLink(CreateLink(FoundItem));
            Stream^.WriteString(L);
            WriteEndOfLink;
          end
        else WriteString(L);
      end
    else
      begin
        if (S[I] = ':')
        then FoundColon := true
        else
        if (S[I] = ';')
        then FoundColon := false;
        WriteChar(S[I]);
        Inc(I);
      end;
  end;
  Stream^.WriteString('</code>');
end;

procedure THTMLDocGenerator.WriteComment(const S: String);
begin
  Stream^.WriteString('<!--');
  WriteString(S);
  Stream^.WriteLine('-->');
end;

procedure THTMLDocGenerator.WriteDates(HL: Byte; Created, LastMod: PString);
begin
  if Assigned(Created)
  then
    begin
      WriteHeading(HL, GetString(TEXT_CREATED));
      WriteStartOfParagraph;
      WriteString(Created^);
      WriteEndOfParagraph;
    end;
  if Assigned(LastMod)
  then
    begin
      WriteHeading(HL, GetString(TEXT_LASTMOD));
      WriteStartOfParagraph;
      WriteLine(LastMod^);
      WriteEndOfParagraph;
    end;
end;

procedure THTMLDocGenerator.WriteDocumentation;
begin
  WriteUnits(1);
  CloseStream;
  WriteOverviewFiles;
end;

procedure THTMLDocGenerator.WriteDocumentHeadline;
var
  I: Byte;
begin
  Stream^.WriteLine('<table cellspacing=1 cellpadding=3 width=100%>');
  Stream^.WriteLine('<tr>');
  for I := 0 to NUM_OVERVIEW_FILES - 1 do
  begin
    Stream^.WriteString('<td bgcolor="' + HTML_HEADER_BACKGROUND_COLOR + '">' +
      '<a href="' + OverviewFilenames[I] + GetFileExtension +
      '"><center>');
    WriteString(GetString(TEXT_OVERVIEW_UNITS + I));
    Stream^.WriteLine('</center></a></td>');
  end;
  Stream^.WriteLine('</tr>');
  Stream^.WriteLine('</table>');
end;

procedure THTMLDocGenerator.WriteEmptyCell;
begin
  Stream^.WriteString('&nbsp;');
end;

procedure THTMLDocGenerator.WriteEndOfDocument;
begin
  Stream^.WriteLine('</body>');
  Stream^.WriteLine('</html>');
end;

procedure THTMLDocGenerator.WriteEndOfAnchor;
begin
  Stream^.WriteString('</a>');
end;

procedure THTMLDocGenerator.WriteEndOfLink;
begin
  Stream^.WriteString('</a>');
end;

procedure THTMLDocGenerator.WriteEndOfParagraph;
begin
  Stream^.WriteLine('</p>');
end;

procedure THTMLDocGenerator.WriteEndOfTableCell;
begin
  Stream^.WriteLine('</td>');
end;

procedure THTMLDocGenerator.WriteEndOfTable;
begin
  Stream^.WriteLine('</table>');
end;

procedure THTMLDocGenerator.WriteEndOfTableRow;
begin
  Stream^.WriteLine('</tr>');
end;

procedure THTMLDocGenerator.WriteFuncsProcs(HL: Byte; Methods: Boolean;
  FuncsProcs: PMethodCollection);
var
  I: LongInt;
  J: LongInt;
  P: PMethod;
begin
  if Methods
  then WriteHeading(HL, GetString(TEXT_METHODS))
  else WriteHeading(HL, GetString(TEXT_FUNCTIONS_AND_PROCEDURES));
  if (not Assigned(FuncsProcs)) or (FuncsProcs^.Count < 1)
  then
    begin
      WriteStartOfParagraph;
      WriteString(GetString(TEXT_NONE));
      WriteEndOfParagraph;
      exit;
    end;
  for I := 0 to 1 do
  begin
    if (I = 0)
    then
      begin
        WriteHeading(HL + 1, GetString(TEXT_OVERVIEW));
        WriteStartOfTable1Column('');
      end
    else WriteHeading(HL + 1, GetString(TEXT_DESCRIPTION));
    for J := 0 to FuncsProcs^.Count - 1 do
    begin
      P := FuncsProcs^.At(J);
      if (I = 0)
      then
        begin
          WriteStartOfTableRow;
          WriteStartOfTableCell;
          WriteStartOfLink(P^.FullLink);
          Stream^.WriteString('<code>');
          WriteString(P^.FullDeclaration);
          Stream^.WriteString('</code>');
          WriteEndOfLink;
          WriteEndOfTableCell;
          WriteEndOfTableRow;
        end
      else
        begin
          WriteLine('');
          WriteStartOfParaGraph;
          WriteStartOfTable1Column('');
          WriteStartOfTableRow;
          WriteStartOfAnchor(AnchorToString(P^.AnchorNumber));
          WriteStartOfTableCell;
          WriteCodeWithLinks(P, P^.FullDeclaration);
          {Stream^.WriteString('<code>');
          WriteString(P^.FullDeclaration);
          Stream^.WriteString('</code>');}
          WriteEndOfTableCell;
          WriteEndOfAnchor;
          WriteEndOfTableRow;
          WriteEndOfTable;
          WriteEndOfParagraph;

          WriteStartOfParagraph;
          WriteItemDescription(P);
          WriteEndOfParagraph;
        end;
    end;
    if (I = 0)
    then WriteEndOfTable;
  end;
end;

procedure THTMLDocGenerator.WriteHeading(I: Byte; const S: String);
var
  C: Char;
begin
  if (I < 1)
  then I := 1;
  if (I > 6)
  then
    begin
      PrintLn(2, 'Warning: HTML generator cannot write headlines of level' +
              ' 7 or greater; will use 6 instead!');
      I := 6;
    end;
  C := Chr(I + 48);
  Stream^.WriteString('<font face="Helvetica, Arial"><h' + C + '>');
  WriteString(S);
  Stream^.WriteLine('</h' + C + '></font>');
end;

procedure THTMLDocGenerator.WriteItemDescription(Item: PItem);
var
  E: Boolean;
begin
  if (not Assigned(Item))
  then exit;
  E := false;
  if (not Assigned(Item^.Description)) and
    (not Assigned(Item^.DetailedDescription))
  then
    begin
      E := true;
    end
  else
    begin
      if Assigned(Item^.DetailedDescription) and
        (Item^.DetailedDescription^.Content > 0)
      then WriteText(Item^.DetailedDescription)
      else
        begin
          if Assigned(Item^.Description) and
            (Item^.Description^.Content > 0)
          then WriteText(Item^.Description)
          else E := true;
        end;
    end;
  if E
  then Stream^.WriteString('&nbsp;');
end;

procedure THTMLDocGenerator.WriteItems(Order: Byte; Heading: String; I: PItemCollection);
var
  J: LongInt;
  Item: PItem;
begin
  WriteHeading(Order, Heading);
  if (not Assigned(I)) or (I^.Count < 1)
  then
    begin
      Stream^.WriteString('<p>');
      WriteString(GetString(TEXT_NONE));
      Stream^.WriteString('</p>');
      exit;
    end;
  Stream^.WriteString('<table border=0 width=100%>');
  Stream^.WriteString('<tr bgcolor="#' + HTML_HEADER_BACKGROUND_COLOR + '">');
  Stream^.WriteLine('<th>' + GetString(TEXT_NAME) + '</th><th>' +
    GetString(TEXT_DESCRIPTION) + '</th></tr>');
  J := 0;
  repeat
    Item := I^.At(J);
    WriteStartOfTableRow;
    WriteStartOfAnchor('link' + Numbers.IntToString(Item^.AnchorNumber));
    WriteStartOfTableCell;
    WriteString(Item^.Name);
    WriteEndOfTableCell;
    WriteStartOfTableCell;
    WriteItemDescription(Item);
    WriteEndOfTableCell;
    WriteEndOfAnchor;
    WriteEndOfTableRow;
    Inc(J);
  until (J = I^.Count);
  Stream^.WriteString('</table>');
end;

procedure THTMLDocGenerator.WriteOverviewFiles;
var
  C1: PItemCollection;
  C2: PItemCollection;
  C3: PItemCollection;
  I: LongInt;
  Item: PItem;
  J: LongInt;
  PU: PUnit;
begin
  WriteUnitOverviewFile;
  if (not Assigned(Units)) or (Units^.Count < 1)
  then exit;
  C3 := New(PItemCollection, Init(16, 16));
  for I := 1 to 5 do
  begin
    if (not CreateStream(OverviewFileNames[I]))
    then
      begin
        PrintLn(1, 'Error: Could not create output file "' +
          OverviewFileNames[I] + '".');
        exit;
      end;
      PrintLn(3, 'Writing overview file ' + Stream^.GetName + '...');
      WriteStartOfDocument(GetString(TEXT_HEADLINE_CIO + I - 1));
      WriteDocumentHeadline;
      WriteHeading(1, GetString(TEXT_HEADLINE_CIO + I - 1));
      C1 := New(PItemCollection, Init(16, 16));
      for J := 0 to Units^.Count - 1 do
      begin
        PU := Units^.At(J);
        case I of
          1: C2 := PU^.CIO;
          2: C2 := PU^.Types;
          3: C2 := PU^.Variables;
          4: C2 := PU^.Constants;
          5: C2 := PU^.FuncsProcs;
        end;
        C1^.InsertItems(C2);
      end;
      if Assigned(C1) and (C1^.Count > 0)
      then
        begin
          WriteStartOfTable3Columns(GetString(TEXT_NAME), GetString(TEXT_UNIT),
            GetString(TEXT_DESCRIPTION));
          for J := 0 to C1^.Count - 1 do
          begin
            Item := C1^.At(J);
            WriteStartOfTableRow;
            WriteStartOfTableCell;
            WriteStartOfLink(Item^.FullLink);
            WriteString(Item^.Name);
            WriteEndOfLink;
            WriteEndOfTableCell;
            WriteStartOfTableCell;
            WriteStartOfLink(Item^.MyUnit^.FullLink);
            WriteString(Item^.MyUnit^.Name);
            WriteEndOfLink;
            WriteEndOfTableCell;
            WriteStartOfTableCell;
            WriteItemDescription(Item);
            WriteEndOfTableCell;
            WriteEndOfTableRow;
          end;
          WriteEndOfTable;
        end
      else
        begin
          WriteStartOfParagraph;
          WriteString(GetString(TEXT_NONE));
          WriteEndOfParagraph;
        end;
    C3^.InsertItems(C1);
    Dispose(C1, Done);
    WriteAppInfo;
    WriteEndOfDocument;
    CloseStream;
  end;
  if (not CreateStream(OverviewFileNames[6]))
  then
    begin
      PrintLn(1, 'Error: Could not create overview output file "' +
        OverviewFileNames[6] + '".');
      exit;
    end;
  PrintLn(3, 'Writing overview file ' + Stream^.GetName + '...');
  WriteStartOfDocument(GetString(TEXT_HEADLINE_IDENTIFIERS));
  WriteDocumentHeadline;
  WriteHeading(1, GetString(TEXT_HEADLINE_IDENTIFIERS));
  WriteStartOfTable3Columns(GetString(TEXT_NAME),
    GetString(TEXT_UNIT), GetString(TEXT_DESCRIPTION));
  for I := 0 to C3^.Count - 1 do
  begin
    Item := C3^.At(I);
    WriteStartOfTableRow;

    WriteStartOfTableCell;
    WriteStartOfLink(Item^.FullLink);
    WriteString(Item^.Name);
    WriteEndOfLink;
    WriteEndOfTableCell;

    WriteStartOfTableCell;
    WriteStartOfLink(Item^.MyUnit^.FullLink);
    WriteString(Item^.MyUnit^.Name);
    WriteEndOfLink;
    WriteEndOfTableCell;

    WriteStartOfTableCell;
    WriteItemDescription(Item);
    WriteEndOfTableCell;

    WriteEndOfTableRow;
  end;
  WriteEndOfTable;
  WriteAppInfo;
  WriteEndOfDocument;
  CloseStream;
end;

procedure THTMLDocGenerator.WriteParagraph(HL: Byte; S: String; T: PText);
begin
  if (not Assigned(T)) or (T^.Content < 1)
  then exit;
  WriteHeading(HL, S);
  Stream^.WriteLine('<p>');
  Stream^.WriteBytes(T^.Content, T^.Data^);
  Stream^.WriteLine('</p>');
end;

procedure THTMLDocGenerator.WriteProperties(HL: Byte; P: PPropertyCollection);
var
  I: LongInt;
  J: LongInt;
  Prop: PProperty;
begin
  WriteHeading(HL, GetString(TEXT_PROPERTIES));
  if (not Assigned(P)) or (P^.Count < 1) then
  begin
    Stream^.WriteLine('<p>');
    WriteString(GetString(TEXT_NONE));
    Stream^.WriteLine('</p>');
    exit;
  end;
  for I := 0 to 1 do
  begin
    if (I = 0)
    then WriteHeading(HL + 1, GetString(TEXT_OVERVIEW))
    else WriteHeading(HL + 1, GetString(TEXT_DESCRIPTION));
    for J := 0 to P^.Count - 1 do
    begin
      Prop := P^.At(J);
      WriteStartOfTable1Column('');
      WriteStartOfTableRow;
      WriteStartOfTableCell;
      if (I = 0)
      then
        begin
          Stream^.WriteString('<code>');
          Stream^.WriteString(CreateReferencedLink(Prop^.FullDeclaration,
            Prop^.FullLink));
          Stream^.WriteString('</code>');
        end
      else
        begin
          WriteStartOfAnchor(AnchorToString(Prop^.AnchorNumber));
          {WriteString(Prop^.FullDeclaration);}
          WriteCodeWithLinks(Prop, Prop^.FullDeclaration);
          {Stream^.WriteString(S);}
          WriteEndOfAnchor;
        end;
      WriteEndOfTableCell;
      WriteEndOfTableRow;
      WriteEndOfTable;
      if (I = 1)
      then
        begin
          WriteStartOfParagraph;
          WriteItemDescription(Prop);
          WriteEndOfParagraph;
        end;
    end;
  end;
end;

procedure THTMLDocGenerator.WriteStartOfAnchor(const Name: String);
begin
  Stream^.WriteString('<a name="' + Name + '">');
end;

procedure THTMLDocGenerator.WriteStartOfDocument(Name: String);
begin
  Stream^.WriteLine('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN"');
  Stream^.WriteLine('"http://www.w3.org/TR/REC-html40/loose.dtd">');
  Stream^.WriteLine('<html>');
  Stream^.WriteLine('<head>');
  Stream^.WriteLine('<title>');
  WriteString(Name);
  Stream^.WriteLine('</title>');
  Stream^.WriteLine('</head>');
  { this warning tells people who view the HTML source code that it was
    generated automatically and that it's of no use to edit the code as
    it's likely to be overwritten soon }
  WriteComment(GetString(TEXT_WARNING_AUTOMATIC));
  Stream^.WriteLine('<body bgcolor="#ffffff" text="#000000" link="#0000ff" vlink="#800080" alink="#FF0000">');
end;

procedure THTMLDocGenerator.WriteStartOfLink(const Name: String);
begin
  Stream^.WriteString('<a href="' + Name + '">');
end;

procedure THTMLDocGenerator.WriteStartOfParagraph;
begin
  Stream^.WriteString('<p>');
end;

procedure THTMLDocGenerator.WriteStartOfTable1Column(T: String);
begin
  Stream^.WriteLine('<table cellspacing=1 cellpadding=3 width=100%>');
  { T is not used - looks ugly }
end;

procedure THTMLDocGenerator.WriteStartOfTable2Columns(T1, T2: String);
begin
  Stream^.WriteLine('<table cellspacing=1 cellpadding=3 width=100%>');
  Stream^.WriteString('<tr bgcolor="#' + HTML_HEADER_BACKGROUND_COLOR +
    '"><th>');
  WriteString(T1);
  Stream^.WriteString('</th><th>');
  WriteString(T2);
  Stream^.WriteLine('</th></tr>');
end;

procedure THTMLDocGenerator.WriteStartOfTable3Columns(T1, T2, T3: String);
begin
  Stream^.WriteLine('<table cellspacing=1 cellpadding=3 width=100%>');
  Stream^.WriteString('<tr bgcolor="#' +
    HTML_HEADER_BACKGROUND_COLOR + '"><th>');
  WriteString(T1);
  Stream^.WriteString('</th><th>');
  WriteString(T2);
  Stream^.WriteString('</th><th>');
  WriteString(T3);
  Stream^.WriteLine('</th></tr>');
end;

procedure THTMLDocGenerator.WriteStartOfTableCell;
begin
  Stream^.WriteString('<td bgcolor="#' + HTML_ROW_BACKGROUND_COLOR + '">');
end;

procedure THTMLDocGenerator.WriteStartOfTableRow;
begin
  Stream^.WriteString('<tr>');
end;

procedure THTMLDocGenerator.WriteUnit(HL: Byte; U: PUnit);
var
  D: PText;
begin
  if (not Assigned(U)) then
  begin
    PrintLn(1, 'Error in THTMLDocGenerator.WriteUnit: Unit variable has not been initialized.');
    exit;
  end;
  if (not CreateStream(U^.Name)) then
  begin
    PrintLn(1, 'Error: Could not create HTML unit doc file for unit ' + U^.Name);
    exit;
  end;
  PrintLn(3, 'Writing documentation for unit "' + U^.Name + '"...');
  WriteStartOfDocument(U^.Name);
  WriteDocumentHeadline;
  WriteHeading(HL, GetString(TEXT_UNIT) + ' ' + U^.Name);
  if (Assigned(U^.DetailedDescription))
  then D := U^.DetailedDescription
  else D := U^.Description;
  WriteUnitDescription(HL + 1, D);
  WriteConstants(HL + 1, U^.Constants);
  WriteVariables(HL + 1, U^.Variables);
  WriteTypes(HL + 1, U^.Types);
  WriteFuncsProcs(HL + 1, false, U^.FuncsProcs);
  WriteCIOSummary(HL + 1, U^.CIO);
  WriteAuthors(HL + 1, U^.Authors);
  WriteDates(HL + 1, U^.Created, U^.LastMod);
  WriteAppInfo;
  WriteEndOfDocument;
  CloseStream;
  WriteCIOs(HL, U^.CIO);
end;

procedure THTMLDocGenerator.WriteUnitDescription(HL: Byte; D: PText);
begin
  WriteParagraph(HL, GetString(TEXT_DESCRIPTION), D);
end;

procedure THTMLDocGenerator.WriteUnitOverviewFile;
var
  C: PItemCollection;
  Item: PItem;
  J: LongInt;
begin
  C := Units;
  if (not CreateStream(OverviewFileNames[0]))
  then
    begin
      PrintLn(1, 'Error: Could not create overview output file "' + OverviewFileNames[0] + '".');
      exit;
    end;
  PrintLn(3, 'Writing unit overview file "' + Stream^.GetName + '"...');
  WriteStartOfDocument(GetString(TEXT_HEADLINE_UNITS));
  WriteDocumentHeadline;
  WriteHeading(1, GetString(TEXT_HEADLINE_UNITS));
  if Assigned(C) and (C^.Count > 0)
  then
    begin
      WriteStartOfTable2Columns(GetString(TEXT_NAME),
      GetString(TEXT_DESCRIPTION));
      for J := 0 to C^.Count - 1 do
      begin
        Item := C^.At(J);
        WriteStartOfTableRow;
        WriteStartOfTableCell;
        WriteStartOfLink(Item^.FullLink);
        WriteString(Item^.Name);
        WriteEndOfLink;
        WriteEndOfTableCell;
        WriteStartOfTableCell;
        WriteItemDescription(Item);
        WriteEndOfTableCell;
        WriteEndOfTableRow;
      end;
      WriteEndOfTable;
    end;
  WriteAppInfo;
  WriteEndOfDocument;
  CloseStream;
end;

begin
end.
