{
@abstract(provides all the parsing functionality of pasdoc)
@author(Marco Schmidt (marcoschmidt@geocities.com))
@lastmod(13 Apr 2000)

Parsing implements most of the functionality of the pasdoc program.

It provides the @link(TParser) object, which scans the command line parameters
for file names and switches and then starts collecting information from those
files, issueing warnings to standard out if necessary.
}

unit parsing;

interface

{$I platform.inc}

uses
  items, objects, scanning, streams, texts, tokenizi;

const
  { indicates that method is a constructor }
  METHOD_CONSTRUCTOR = 0;
  { indicates that method is a destructor }
  METHOD_DESTRUCTOR = 1;
  { indicates that method is a function or procedure and not a constructor
    or destructor }
  METHOD_FUNCTION_PROCEDURE = 2;

type
  { pointer to @link(TParser) }
  PParser = ^TParser;
  { Parser object that will process a complete unit file and all of its
    include files, regarding directives.
  Will normally be created used @link(Init) with an input stream
  and a list of directives, then does the parsing work when its
  @link(ParseUnit) method is called.
  If no errors appear, should return a @link(PUnit) object with
  all information on the unit.
  Otherwise a description of the error should be found in
  @link(ErrorMessage). }
  TParser = object(TObject)
    { Stores description of error whenever something went wrong. }
    ErrorMessage: String;
    { Last comment found in input or nil if no comment available.
    Will be modified by @link(GetLastComment). }
    LastCommentToken: PToken;
    { The underlying scanner object. }
    Scanner: PScanner;
    { Create a parser, initialize the scanner with input stream S.
      All strings in SD are defined compiler directives. }
    constructor Init(InputStream: PInputStream; SD: PStringCollection; IFP: PCollection);
    { Release all dynamically allocated memory. }
    destructor Done; virtual;
    { Will write last comment that was found in input to T. If there was none,
      T will be set to nil. }
    procedure GetLastComment(var T: PText);
    { Get next token T from scanner that is neither whitespace nor comment.
      Return true on success. }
    function GetNextNonWCToken(var T: PToken): Boolean;
    function ParseArguments(var A: String): Boolean;
    { Parses a constructor, a destructor, a function or a procedure.
      Resulting PMethod item will be returned in M.
      CS may contain the 'class' keyword - its exact spelling is taken from
      this variable.
      CDFP contains the keyword constructor, destructor, function or procedure
      in the exact spelling as it was found in input.
      Key contains one of the KEY_xxx constants for the What field of the
      resulting method object.
      D may contain a description or nil. }
    function ParseCDFP(var M: PMethod; CS, CDFPS: String; Key: Integer; D: PText): Boolean;
    { Parses a class, an interface or an object.
      U is the unit this item will be added to on success.
      N is the name of this item.
      CIOType describes if item is class, interface or object.
      D may contain a description or nil. }
    function ParseCIO(var U: PUnit; N: String; CIOType: TCIOType; D: PText):
      Boolean;
    { }
    function ParseConstant(var U: PUnit; T: PToken): Boolean;
    function ParseInterfaceSection(var U: PUnit): Boolean;
    function ParseProperty(var P: PProperty): Boolean;
    function ParseType(var U: PUnit; var T: PToken): Boolean;
    function ParseUnit(var U: PUnit): Boolean;
    function ParseUses(var U: PUnit): Boolean;
    function ParseVariables(var U: PUnit; var T: PToken): Boolean;
    function SkipDeclaration: Boolean;
    { Reads tokens and throws them away as long as they are either whitespace
      or comments.
      Returns true on success, false if there were any errors. }
    function SkipWhitespaceAndComments: Boolean;
  end;

implementation

uses
  arrays, msg, numbers;

{ TParser }

constructor TParser.Init(InputStream: PInputStream; SD: PStringCollection; IFP: PCollection);
begin
  inherited Init;
  LastCommentToken := nil;
  Scanner := New(PScanner, Init(InputStream));
  Scanner^.AddDirectives(SD);
  Scanner^.IncludeFilePaths := IFP;
end;

destructor TParser.Done;
begin
  if Assigned(Scanner)
  then Dispose(Scanner, Done);
  inherited Done;
end;

procedure TParser.GetLastComment(var T: PText);
begin
  if Assigned(LastCommentToken) then
  begin
    T := LastCommentToken^.Data;
    LastCommentToken^.Data := nil;
    Dispose(LastCommentToken, Done);
    LastCommentToken := nil;
    { remove comment characters here }
    if (T^.Content >= 1) and (T^.Data^[0] = '{')
    then T^.Delete(0, 1);
    if (T^.Content >= 1) and (T^.Data^[T^.Content - 1] = '}')
    then T^.Delete(T^.Content - 1, 1);
    if (T^.Content >= 2) and (T^.Data^[0] = '(') and (T^.Data^[1] = '*')
    then T^.Delete(0, 2);
    if (T^.Content >= 2) and (T^.Data^[0] = '/') and (T^.Data^[1] = '/')
    then T^.Delete(0, 2);
    if (T^.Content >= 2) and (T^.Data^[T^.Content - 2] = '*') and
     (T^.Data^[T^.Content - 1] = ')')
    then T^.Delete(T^.Content - 2, 2);
  end
  else T := nil;
end;

function TParser.GetNextNonWCToken(var T: PToken): Boolean;
begin
  GetNextNonWCToken := false;
  if (not SkipWhitespaceAndComments)
  then exit;
  GetNextNonWCToken := Scanner^.GetToken(T);
end;

function TParser.ParseArguments(var A: String): Boolean;
var
  Finished: Boolean;
  T: PToken;
begin
  ParseArguments := false;
  Finished := false;
  A := '';
  repeat
    if (not Scanner^.GetToken(T))
    then exit;
    if (T^.MyType = TOK_SYMBOL) and (T^.Info.SymbolType = SYM_RIGHT_PARENTHESIS)
    then Finished := true
    else
    if (T^.MyType = TOK_SYMBOL) and
       ((T^.Info.SymbolType = SYM_COLON) or
        (T^.Info.SymbolType = SYM_COMMA) or
        (T^.Info.SymbolType = SYM_SEMICOLON)) then
    begin
      if (Length(A) > 0) and (A[Length(A)] = ' ')
      then SetLength(A, Length(A) - 1);
      A := A + T^.Data^.GetString;
    end
    else
    if (T^.MyType = TOK_WHITESPACE) then
    begin
      if (Length(A) > 0) and (A[Length(A)] <> ' ')
      then A := A + ' ';
    end
    else
    if (T^.MyType = TOK_COMMENT) or (T^.MyType = TOK_DIRECTIVE) then
    begin
      { ignore }
    end
    else { otherwise copy } A := A + T^.Data^.GetString;
    Dispose(T, Done);
  until Finished;
  ParseArguments := true;
end;

function TParser.ParseCDFP(var M: PMethod; CS, CDFPS: String; Key: Integer; D: PText): Boolean;
const
  SDSet: Set of Byte = [SD_ABSTRACT, SD_ASSEMBLER, SD_CDECL, SD_EXPORT,
    SD_EXTERNAL, SD_FAR, SD_FORWARD, SD_NEAR, SD_OVERLOAD, SD_OVERRIDE,
  SD_STDCALL, SD_VIRTUAL];
var
  Finished: Boolean;
  IsSemicolon: Boolean;
  PL: Integer;
  T: PToken;
begin
  ParseCDFP := false;
  M := New(PMethod, Init);
  if (not Assigned(M)) then
  begin
    ErrorMessage := 'Error, could not create CDFP item.';
    exit;
  end;
  M^.Description := D;
  case Key of
    KEY_CONSTRUCTOR: M^.What := METHOD_CONSTRUCTOR;
    KEY_DESTRUCTOR: M^.What := METHOD_DESTRUCTOR;
    KEY_FUNCTION,
    KEY_PROCEDURE: M^.What := METHOD_FUNCTION_PROCEDURE;
    else
      begin
        PrintLn(1, 'FATAL ERROR: CDFP got invalid key.');
        Halt(1);
      end;
  end;
  { next non-wc token must be the name }
  if (not GetNextNonWCToken(T)) then
  begin
    Dispose(M, Done);
    exit;
  end;
  if (T^.MyType <> TOK_IDENTIFIER) then
  begin
    Dispose(M, Done);
    Dispose(T, Done);
    exit;
  end;
  if (Length(CS) > 0)
  then CS := CS + ' ';
  M^.Name := T^.Data^.GetString;
  PrintLn(5, 'Parsing ' + CDFPS + ' ' + M^.Name);
  M^.FullDeclaration := CS + CDFPS + ' ' + M^.Name;
  Dispose(T, Done);
  { copy tokens until first semicolon with parenthesis level zero }
  PL := 0;
  repeat
    if (not Scanner^.GetToken(T)) then
    begin
      Dispose(M, Done);
      exit;
    end;
    if (T^.MyType = TOK_COMMENT)
    then
    else
    if (T^.MyType = TOK_WHITESPACE)    then
    begin
      if (M^.FullDeclaration[Length(M^.FullDeclaration)] <> ' ')
      then M^.FullDeclaration := M^.FullDeclaration + ' ';
    end
    else
    begin
      M^.FullDeclaration := M^.FullDeclaration + T^.Data^.GetString;
    end;
    if (T^.MyType = TOK_SYMBOL) and (T^.Info.SymbolType = SYM_LEFT_PARENTHESIS)
    then Inc(PL);
    if (T^.MyType = TOK_SYMBOL) and (T^.Info.SymbolType = SYM_RIGHT_PARENTHESIS)
    then Dec(PL);
    IsSemicolon := (T^.MyType = TOK_SYMBOL) and (T^.Info.SymbolType = SYM_SEMICOLON);
    Dispose(T, Done);
  until IsSemicolon and (PL = 0);
  { first get non-WC token - if it is not an identifier in SD_SET put it back
    into stream and leave; otherwise copy tokens until semicolon }
  Finished := false;
  repeat
    if (not GetNextNonWCToken(T)) then
    begin
      Dispose(M, Done);
      exit;
    end;
    if (T^.MyType <> TOK_IDENTIFIER) then
    begin
      Scanner^.UnGetToken(T);
      Break;
    end;
    CS := T^.Data^.GetString;
    StringToUpper(CS, CS);
    PL := StandardDirectives^.IndexOf(@CS);
    if (PL < 0) or (not (PL in SDSet)) then
    begin
      Scanner^.UnGetToken(T);
      Break;
    end;
    M^.FullDeclaration := M^.FullDeclaration + ' ' + T^.Data^.GetString;
    Dispose(T, Done);
    repeat
      if (not GetNextNonWCToken(T)) then
      begin
        Dispose(M, Done);
        exit;
      end;
      IsSemicolon := (T^.MyType = TOK_SYMBOL) and (T^.Info.SymbolType = SYM_SEMICOLON);
      if (not IsSemicolon)
      then M^.FullDeclaration := M^.FullDeclaration + ' ';
      M^.FullDeclaration := M^.FullDeclaration + T^.Data^.GetString;
    until IsSemicolon;
  until Finished;
  ParseCDFP := true;
end;

function TParser.ParseCIO(var U: PUnit; N: String; CIOType: TCIOType;
  D: PText): Boolean;
var
  CS: String;
  CSFound: Boolean;
  F: PItem;
  Finished: Boolean;
  I: PCIO;
  Ind: Integer;
  M: PMethod;
  P: PProperty;
  S: String;
  State: Byte;
  T: PToken;
begin
  ParseCIO := false;
  PrintLn(5, 'Parsing class/interface/object "' + N + '"');
  if (not GetNextNonWCToken(T))
  then exit;
  I := New(PCIO, Init);
  I^.Name := N;
  I^.Description := D;
  I^.MyType := CIOType;
  { get all ancestors; remember, this could look like
    TNewClass = class ( Classes.TClass, MyClasses.TFunkyClass,
                MoreClasses.YAC) ... end;
    all class ancestors are supposed to be included in the docs!
  }
  if (T^.IsSymbol(SYM_LEFT_PARENTHESIS)) then
  begin
    { optional ancestor introduced by ( }
    Dispose(T, Done);
    Finished := false;
    I^.Ancestors := New(PStringCollection, Init(1, 1));
    { outer repeat loop: one ancestor per pass }
    repeat
      if (not GetNextNonWCToken(T))
      then exit;
      if (T^.MyType = TOK_IDENTIFIER) then
      begin { an ancestor }
        S := T^.Data^.GetString;
        Dispose(T, Done);
        { inner repeat loop: one part of the ancestor per name }
        repeat
          if (not Scanner^.GetToken(T)) then
          begin
            exit;
          end;
          if (not T^.IsSymbol(SYM_PERIOD)) then
          begin
            Scanner^.UnGetToken(T);
            Break; { leave inner repeat loop }
          end;
          Dispose(T, Done);
          S := S + '.';
          if (not Scanner^.GetToken(T)) or (T^.MyType <> TOK_IDENTIFIER) then
          begin
            ErrorMessage := Scanner^.GetStreamInfo + ': Error - expected ' +
              'class, object or interface in ancestor declaration.';
            exit;
          end;
          S := S + T^.Data^.GetString;
        until false;
        I^.Ancestors^.Insert(Objects.NewStr(S));
      end
      else
      if (T^.IsSymbol(SYM_COMMA)) then
      { comma, separating two ancestors }
      begin
        Dispose(T, Done)
      end
      else
      begin
        Finished := T^.IsSymbol(SYM_RIGHT_PARENTHESIS);
        Dispose(T, Done);
        if (not Finished) then
        begin
          ErrorMessage := Scanner^.GetStreamInfo + ': Error - ")" expected.';
          exit;
        end;
      end;
    until Finished;
  end
  else
  { optional interface ID introduced by [ }
  if (T^.IsSymbol(SYM_LEFT_BRACKET)) then
  begin
    Dispose(T, Done);
    { for the time being, we throw away the ID itself }
    if (not GetNextNonWCToken(T)) then
    begin
      exit;
    end;
    if (T^.MyType <> TOK_STRING) then
    begin
      ErrorMessage := Scanner^.GetStreamInfo +
        ': Error - literal string as interface ID expected.';
      exit;
    end;
    if (not GetNextNonWCToken(T)) then
    begin
      exit;
    end;
    if (not T^.IsSymbol(SYM_RIGHT_BRACKET)) then
    begin
      ErrorMessage := Scanner^.GetStreamInfo + ': Error - "]" expected.';
      exit;
    end;
  end
  else Scanner^.UnGetToken(T);
  { now collect methods, fields and properties }
  CS := '';
  State := STATE_PUBLIC;
  Finished := false;
  repeat
    CSFound := false;
    if (not GetNextNonWCToken(T)) then
    begin
      Dispose(I, Done);
      exit;
    end;
    if (T^.IsSymbol(SYM_SEMICOLON)) then
    begin
      { a forward declaration of type "name = class(ancestor);" - this is
        ignored, we simply leave indicating no error }
      Dispose(I, Done);
      Dispose(T, Done);
      ParseCIO := true;
      exit;
    end
    else
    if (T^.MyType = TOK_RESERVED)
    then
      case T^.Info.ReservedKey of
        KEY_CLASS:
        begin
          CS := T^.Data^.GetString;
          CSFound := true;
        end;
        KEY_CONSTRUCTOR,
        KEY_DESTRUCTOR,
        KEY_FUNCTION,
        KEY_PROCEDURE:
        begin
          GetLastComment(D);
          if (not ParseCDFP(M, CS, T^.Data^.GetString, T^.Info.ReservedKey, D)) then
          begin
            Dispose(I, Done);
            Dispose(T, Done);
            exit;
          end;
          M^.State := State;
          M^.InsertMethod(M, I^.Methods);
        end;
        KEY_END: Finished := true;
        KEY_PROPERTY:
        begin
          if (not ParseProperty(P)) then
          begin
            exit;
          end;
          P^.State := State;
          P^.InsertProperty(P, I^.Properties);
        end;
        else
        begin
          ErrorMessage := Scanner^.GetStreamInfo +
            ': Error, unexpected reserved keyword "' +
          KeyWordArray[T^.Info.ReservedKey] + '"';
            Dispose(I, Done);
            Dispose(T, Done);
            exit;
        end;
      end
    else
    if (T^.MyType = TOK_IDENTIFIER)
    then
      begin
        CS := T^.Data^.GetString;
        StringToUpper(CS, CS);
        Ind := StandardDirectives^.IndexOf(@CS);
        case Ind of
          SD_DEFAULT:
          begin
            if (not SkipDeclaration) then
            begin
              ErrorMessage := 'Could not skip declaration after default property';
              exit;
            end;
            PrintLn(5, 'Skipped default property keyword.');
          end;
          SD_PUBLIC: State := STATE_PUBLIC;
          SD_PUBLISHED: State := STATE_PUBLISHED;
          SD_PRIVATE: State := STATE_PRIVATE;
          SD_PROTECTED: State := STATE_PROTECTED;
          else Ind := -1;
        end;
        if (Ind = -1)
        then
          begin
            F := New(PItem, Init);
            if (not Assigned(F)) then
              begin
              	ErrorMessage := 'Could not create field object.';
                exit;
              end;
            F^.Name := T^.Data^.GetString;
            F^.State := State;
            GetLastComment(F^.Description);
            if (not SkipDeclaration)
            then
              begin
                exit;
              end;
            F^.InsertItem(F, I^.Fields);
          end;
      end;
    if (not CSFound)
    then CS := '';
    Dispose(T, Done);
  until Finished;
  if (not GetNextNonWCToken(T)) or (not T^.IsSymbol(SYM_SEMICOLON))
  then
    begin
      ErrorMessage := Scanner^.GetStreamInfo +
        ' semicolon at the end of class/obj/interf. expected.';
      Dispose(I, Done);
      exit;
    end;
  Dispose(T, Done);
  U^.AddCIO(I);
  ParseCIO := true;
end;

function TParser.ParseConstant(var U: PUnit; T: PToken): Boolean;
var
  I: PItem;
begin
  ParseConstant := false;
  I := New(PItem, Init);
  if (not Assigned(T)) or (not Assigned(T^.Data)) then
  begin
  	PrintLn(1, 'INTERNAL ERROR in ParseConstant - token not initialized.');
  	exit;
  end;
  I^.Name := T^.Data^.GetString;
  PrintLn(5, 'Parsing constant ' + I^.Name);
  GetLastComment(I^.Description);
  if SkipDeclaration then
  begin
    U^.AddConstant(I);
    ParseConstant := true;
  end
  else
  begin
    ErrorMessage := 'Could not skip declaration of constant ' + I^.Name;
  end;
end;

function TParser.ParseInterfaceSection(var U: PUnit): Boolean;
const
  MODE_UNDEFINED = 0;
  MODE_CONST = 1;
  MODE_TYPE = 2;
  MODE_VAR = 3;
var
  D: PText;
  Finished: Boolean;
  Mode: Integer;
  M: PMethod;
  S: String;
  T: PToken;
begin
  PrintLn(4, 'Entering interface section of unit ' + U^.Name);
  ParseInterfaceSection := false;
  Finished := false;
  Mode := MODE_UNDEFINED;
  repeat
    if (not GetNextNonWCToken(T))
    then
      begin
        PrintLn(2, Scanner^.GetStreamInfo +
          ': Error, could not get next non-whitespace, non-comment token');
      exit;
      end;
    case T^.MyType of
      TOK_IDENTIFIER:
      begin
        S := T^.Data^.GetString;
        case Mode of
          MODE_CONST:
            if (not ParseConstant(U, T))
            then exit;
          MODE_TYPE:
            if (not ParseType(U, T))
            then exit;
          MODE_VAR:
            if (not ParseVariables(U, T))
            then exit;
          else
            begin
              ErrorMessage := Scanner^.GetStreamInfo + ': Error, unexpected '+
                'identifier "' + T^.Data^.GetString + '"';
              exit;
            end;
        end;
      end;
      TOK_RESERVED:
      begin
        case T^.Info.ReservedKey of
          KEY_RESOURCESTRING,
          KEY_CONST:
            Mode := MODE_CONST;
          KEY_FUNCTION,
          KEY_PROCEDURE:
          begin
            GetLastComment(D);
            if (not ParseCDFP(M, '', T^.Data^.GetString,
                T^.Info.ReservedKey, D))
            then
              begin
                exit;
              end;
            M^.InsertMethod(M, U^.FuncsProcs);
            Mode := MODE_UNDEFINED;
          end;
          KEY_IMPLEMENTATION:
            Finished := true;
          KEY_TYPE:
            Mode := MODE_TYPE;
          KEY_USES:
            if (not ParseUses(U))
            then exit;
          KEY_THREADVAR,
          KEY_VAR:
            Mode := MODE_VAR;
          else
            begin
              ErrorMessage := Scanner^.GetStreamInfo +
                ': Error, unexpected keyword ' + T^.Data^.GetString;
              exit;
            end;
        end;
      end;
    end;
    if Assigned(T)
    then Dispose(T, Done);
  until Finished;
  ParseInterfaceSection := true;
end;

function TParser.ParseProperty(var P: PProperty): Boolean;
var
  Finished: Boolean;
  T: PToken;
begin
  ParseProperty := false;
  if (not GetNextNonWCToken(T))
  then exit;
  if (T^.MyType <> TOK_IDENTIFIER) then
  begin
    Dispose(T, Done);
    ErrorMessage := Scanner^.GetStreamInfo +
      ': expected identifier as property name.';
    exit;
  end;
  P := New(PProperty, Init);
  if (not Assigned(P)) then
  begin
    ErrorMessage := 'Could not create property item.';
    exit;
  end;
  P^.Name := T^.Data^.GetString;
  PrintLn(5, 'Parsing property ' + P^.Name);
  P^.IndexDecl := '';
  P^.PropType := '';
  Dispose(T, Done);
  GetLastComment(P^.Description);
  if (not GetNextNonWCToken(T))
  then exit;
  { get index }
  if (T^.IsSymbol(SYM_LEFT_BRACKET)) then
  begin
    Dispose(T, Done);
    P^.IndexDecl := '[';
    repeat
      if (not Scanner^.GetToken(T)) then
      begin
        PrintLn(2, Scanner^.GetStreamInfo + ': Error, could not parse property.');
        exit;
      end;
      if (T^.MyType <> TOK_COMMENT) and (T^.MyType <> TOK_DIRECTIVE)
      then P^.IndexDecl := P^.IndexDecl + T^.Data^.GetString;
      Finished := T^.IsSymbol(SYM_RIGHT_BRACKET);
      Dispose(T, Done);
    until Finished;
    { next nonwc token should be the colon }
    if (not GetNextNonWCToken(T))
    then exit;
  end
  else
    begin
      if (T^.IsSymbol(SYM_SEMICOLON))
      then
        begin
          P^.FullDeclaration := P^.Name + P^.IndexDecl + ': ' + P^.PropType;
          Dispose(T, Done);
          ParseProperty := true;
          exit;
        end;
    end;
  { now there should be a colon followed by the type }
{  if (not T^.IsSymbol(SYM_COLON))
  then
    begin
      ErrorMessage := Scanner^.GetStreamInfo + ': expected ":" character.';
      Dispose(P, Done);
      P := nil;
      Dispose(T, Done);
      exit;
    end;
  Dispose(T, Done);}
  { get property type }
{  if (not GetNextNonWCToken(T))
  then exit;
  if (T^.MyType <> TOK_IDENTIFIER) and (T^.MyType <> TOK_RESERVED)
  then
    begin
      PrintLn(2, 'Property error, identifier expected, found ' + TokenTypeNames[T^.MyType]);
      ErrorMessage := Scanner^.GetStreamInfo + ' identifier expected';
      exit;
    end;
  P^.PropType := T^.Data^.GetString;
  Dispose(T, Done);}
  { simply skipping the rest of declaration }
  if (not SkipDeclaration) then
  begin
      PrintLn(2, 'Error: could not skip rest of declaratino');
      exit;
    end;
  P^.FullDeclaration := P^.Name + P^.IndexDecl + ': ' + P^.PropType;
  ParseProperty := true;
end;

{
  TYPENAME =
    class of ... ;               => "normal" type
    class ( ANCESTOR<S> )        => class
          ANYTHING               => class
    object ( ) end ;                 => object
          ANYTHING
    interface end ;              => interface
}
function TParser.ParseType(var U: PUnit; var T: PToken): Boolean;
var
  D: PText;
  I: PItem;
  N: String;
begin
  ParseType := false;
  N := T^.Data^.GetString;
  PrintLn(5, 'Parsing type "' + N + '"');
  Dispose(T, Done);
  GetLastComment(D);
  if (not GetNextNonWCToken(T))
  then exit;
  if (not T^.IsSymbol(SYM_EQUAL))
  then
    begin
      if (T^.IsSymbol(SYM_SEMICOLON))
      then
        begin
          Dispose(T, Done);
          T := nil;
          ParseType := true;
          exit;
        end;
      ErrorMessage := Scanner^.GetStreamInfo + ': Error, "=" expected.';
      Dispose(T, Done);
      exit;
    end;
  Dispose(T, Done);
  if (not GetNextNonWCToken(T))
  then exit;
  if (T^.MyType = TOK_RESERVED)
  then
    case T^.Info.ReservedKey of
      KEY_CLASS:
      begin
        Dispose(T, Done);
        if (not GetNextNonWCToken(T))
        then exit;
        if (T^.MyType = TOK_RESERVED) and (T^.Info.ReservedKey = KEY_OF)
        then
      begin
        { include "identifier = class of something;" as standard type }
          end
        else
          begin
            Scanner^.UnGetToken(T);
            if (not ParseCIO(U, N, CIO_CLASS, D))
            then exit;
            T := nil;
            ParseType := true;
            exit;
          end;
      end;
      KEY_INTERFACE:
      begin
        if (not ParseCIO(U, N, CIO_INTERFACE, D))
        then exit;
        Dispose(T, Done);
        T := nil;
        ParseType := true;
        exit;
      end;
      KEY_OBJECT:
      begin
        if (not ParseCIO(U, N, CIO_OBJECT, D))
        then exit;
        Dispose(T, Done);
        T := nil;
        ParseType := true;
        exit;
      end;
    end;
  Scanner^.UnGetToken(T);
  T := nil; { so that calling function will not try to dispose of it }
  if (not SkipDeclaration)
  then exit;
  I := New(PItem, Init);
  if (not Assigned(I))
  then
    begin
      ErrorMessage := 'Error, unable to create new item.';
      exit;
    end;
  I^.Name := N;
  I^.Description := D;
  U^.AddType(I);
  ParseType := true;
end;

function TParser.ParseUnit(var U: PUnit): Boolean;
var
  T: PToken;
begin
  ParseUnit := false;
  { get 'unit' keyword }
  if (not GetNextNonWCToken(T))
  then exit;
  if (T^.MyType <> TOK_RESERVED) or (T^.Info.ReservedKey <> KEY_UNIT)
  then
    begin
      ErrorMessage := Scanner^.GetStreamInfo + ': keyword "unit" expected.';
      exit;
    end;
  Dispose(T, Done);
  U := New(PUnit, Init);
  GetLastComment(U^.Description);
  if (not GetNextNonWCToken(T))
  then exit;
  { get unit name identifier }
  if (T^.MyType <> TOK_IDENTIFIER)
  then
    begin
      ErrorMessage := Scanner^.GetStreamInfo + ': identifier (unit name) expected.';
      exit;
    end;
  U^.Name := T^.Data^.GetString;
  { skip semicolon }
  if (not GetNextNonWCToken(T))
  then exit;
  if (not T^.IsSymbol(SYM_SEMICOLON))
  then
    begin
      ErrorMessage := Scanner^.GetStreamInfo + ': semicolon expected.';
      exit;
    end;
  if (not GetNextNonWCToken(T))
  then exit;
  { get 'interface' keyword }
  if (T^.MyType <> TOK_RESERVED) or (T^.Info.ReservedKey <> KEY_INTERFACE)
  then
    begin
      ErrorMessage := Scanner^.GetStreamInfo + ': keyword "INTERFACE" expected.';
      exit;
    end;
  { now parse the interface section of that unit }
  ParseUnit := ParseInterfaceSection(U);
end;

function TParser.ParseUses(var U: PUnit): Boolean;
var
  Finished: Boolean;
  T: PToken;
begin
  ParseUses := false;
  if (not Assigned(U^.UsesUnits))
  then U^.UsesUnits := New(PStringCollection, Init(1, 1));
  if (not Assigned(U^.UsesUnits)) then
  begin
  	PrintLn(2, 'Error in TParser.ParseUses: Could not create list of unit names.');
    exit;
  end;
  repeat
    if (not GetNextNonWCToken(T))
    then exit;
    if (T^.MyType <> TOK_IDENTIFIER) then
    begin
      ErrorMessage := Scanner^.GetStreamInfo + ': Error, unit name expected (found ' +
      	T^.GetTypeName + ', ' + T^.Data^.GetString + ')';
      exit;
    end;
    U^.UsesUnits^.Insert(NewStr(T^.Data^.GetString));
    Dispose(T, Done);
    if (not GetNextNonWCToken(T))
    then exit;
    if (T^.MyType <> TOK_SYMBOL) and
       (T^.Info.SymbolType <> SYM_COMMA) and
       (T^.Info.SymbolType <> SYM_SEMICOLON) then
    begin
      ErrorMessage := Scanner^.GetStreamInfo + ': Error, comma or semicolon expected.';
      exit;
    end;
    Finished := (T^.Info.SymbolType = SYM_SEMICOLON);
    Dispose(T, Done);
  until Finished;
  ParseUses := true;
end;

function TParser.ParseVariables(var U: PUnit; var T: PToken): Boolean;
var
  Finished: Boolean;
  FirstLoop: Boolean;
  I: PItem;
begin
  ParseVariables := false;
  FirstLoop := true;
  repeat
    I := New(PItem, Init);
    if (not Assigned(I))
    then
      begin
        ErrorMessage := 'Error: Not enough memory for new item.';
        exit;
      end;
    if FirstLoop
    then
      begin
        I^.Name := T^.Data^.GetString;
        FirstLoop := false;
      end
    else
      begin
        if (not GetNextNonWCToken(T))
        then exit;
        if (T^.MyType <> TOK_IDENTIFIER)
        then
          begin
            ErrorMessage := Scanner^.GetStreamInfo +
              ': Error, identifier expected.';
            exit;
          end;
        I^.Name := T^.Data^.GetString;
      end;
    GetLastComment(I^.Description);
    U^.AddVariable(I);
    Dispose(T, Done);
    if (not GetNextNonWCToken(T))
    then exit;
    if (T^.MyType <> TOK_SYMBOL) or
    ((T^.Info.SymbolType <> SYM_COMMA) and
       (T^.Info.SymbolType <> SYM_COLON))
    then
      begin
        ErrorMessage := Scanner^.GetStreamInfo +
      ': Error, expected comma or colon in var declaration.';
        exit;
      end;
    Finished := (T^.Info.SymbolType = SYM_COLON);
    Dispose(T, Done);
    T := nil;
  until Finished;
  if (not SkipDeclaration)
  then exit;
  ParseVariables := true;
end;

function TParser.SkipDeclaration: Boolean;
var
  EndLevel: Integer;
  IsSemicolon: Boolean;
  PLevel: Integer;
  T: PToken;
begin
  SkipDeclaration := false;
  EndLevel := 0;
  PLevel := 0;
  repeat
    if (not GetNextNonWCToken(T))
    then exit;
    IsSemicolon := (T^.MyType = TOK_SYMBOL) and
    (T^.Info.SymbolType = SYM_SEMICOLON);
    if (T^.MyType = TOK_SYMBOL)
  then
    begin
      if (T^.Info.SymbolType = SYM_LEFT_PARENTHESIS)
        then Inc(PLevel);
      if (T^.Info.SymbolType = SYM_RIGHT_PARENTHESIS)
        then Dec(PLevel);
      end;
    if (T^.MyType = TOK_RESERVED)
  then
    begin
      if (T^.Info.ReservedKey = KEY_END)
        then Dec(EndLevel)
        else
      if (T^.Info.ReservedKey = KEY_RECORD)
        then Inc(EndLevel);
      end;
    Dispose(T, Done);
  until IsSemicolon and (EndLevel = 0) and (PLevel = 0);
  SkipDeclaration := true;
end;

function TParser.SkipWhitespaceAndComments: Boolean;
var
  T: PToken;
begin
  repeat
    if (not Scanner^.PeekToken(T)) then
    begin
      SkipWhitespaceAndComments := false;
      exit;
    end;
    if (T^.MyType = TOK_WHITESPACE) then
    begin
      Scanner^.ConsumeToken;
      Dispose(T, Done);
    end
    else
    if (T^.MyType = TOK_COMMENT) then
    begin
      Scanner^.ConsumeToken;
      if Assigned(LastCommentToken)
      then Dispose(LastCommentToken, Done);
      LastCommentToken := T;
    end
    else
    begin
      SkipWhitespaceAndComments := true;
      exit;
    end;
  until false;
end;

begin
end.
