{
@abstract(provides simplified Pascal tokenizer)
@created(7 Mar 1999)
@lastmod(13 Apr 2000)
@author(Marco Schmidt (marcoschmidt@geocities.com))

The @link(TTokenizer) object creates @link(TToken) objects (tokens) for the
Pascal programming language from a character input stream.
The @link(Scanning) unit does the same (it actually uses this unit's
tokenizer), with the exception that it evaluates compiler directives,
which are comments that start with a dollar sign.
}

unit tokenizi;

interface

uses
  objects, streams, texts;

type
  { enumeration type that provides all types of tokens; each token's name
    starts with TOK_ }
  TTokenType = (TOK_WHITESPACE, TOK_COMMENT, TOK_IDENTIFIER,
    TOK_NUMBER, TOK_STRING, TOK_SYMBOL, TOK_DIRECTIVE, TOK_RESERVED);

const
  TOKEN_TYPE_NAMES: Array[TTokenType] of String[15]=
    ('Whitespace', 'Comment', 'Identifier',
     'Number', 'String', 'Symbol', 'Directive', 'Reserved');

type
  { enumeration type that provides all types of symbols; each
    symbol's name starts with SYM_ }
  TSymbolType = (SYM_PLUS, SYM_MINUS, SYM_ASTERISK, SYM_SLASH, SYM_EQUAL,
    SYM_LESS_THAN, SYM_LESS_THAN_EQUAL, SYM_GREATER_THAN,
	SYM_GREATER_THAN_EQUAL, SYM_LEFT_BRACKET, SYM_RIGHT_BRACKET,
    SYM_COMMA, SYM_LEFT_PARENTHESIS, SYM_RIGHT_PARENTHESIS, SYM_COLON,
	SYM_SEMICOLON, SYM_ROOF, SYM_PERIOD, SYM_AT, SYM_LEFT_BRACE,
	SYM_RIGHT_BRACE, SYM_DOLLAR, SYM_NUMBER, SYM_ASSIGN, SYM_RANGE);
  { a pointer to @link(TToken) }
  PToken = ^TToken;
  { Stores the exact type and additional information on one token.
    Additionally, @link(Data) stores the array of characters }
  TToken = object(TObject)
    { the exact character representation of this token as it was found in the
      input file }
    Data: PText;
    { additional information on this token as a record of variant fields }
    Info: record
      case Integer of
        0: (WhitespaceRows: LongInt);
        2: (SymbolType: TSymbolType);
        3: (ReservedKey: LongInt);
    end;
    { the type of this token as @link(TTokenType) }
    MyType: TTokenType;
    { Create a token of and assign the argument token type to @link(MyType) }
    constructor Init(TT: TTokenType);
    function GetTypeName: String;
    { Returns if argument ST equals @link(MyType) }
    function IsSymbol(ST: TSymbolType): Boolean;
  end;
  { a set of Char values }
  TCharSet = Set of Char;
  { pointer to a @link(TScanner) }
  PTokenizer = ^TTokenizer;
  { @abstract(converts a @link(TInputStream) to a sequence of @link(TToken) objects)
  }
  TTokenizer = object(TObject)
    { if @link(IsCharBuffered) is true, this field contains the buffered
	  character }
    BufferedChar: Char;
    { the last error message }
    ErrorMessage: String;
    { true if end of stream @link(Stream) has been reached, false otherwise }
    EOS: Boolean;
    { if this is true, @link(BufferedChar) contains a buffered character;
      the next call to @link(GetChar) or @link(PeekChar) will return this
      character, not the next in the associated stream @link(Stream) }
    IsCharBuffered: Boolean;
    { current row in stream @link(Stream); useful when giving error messages }
    Row: LongInt;
    { the input stream this tokenizer is working on }
    Stream: PInputStream;
    { Creates a TTokenizer and associates it with given @link(TInputStream). }
    constructor Init(S: PInputStream);
    { Releases all dynamically allocated memory. }
    destructor Done; virtual;
    procedure CheckForDirective(var T: TToken; Offset: Integer);
    procedure ConsumeChar;
    function CreateSymbolToken(ST: TSymbolType; S: String; var T: PToken): Boolean;
    function CreateToken(TT: TTokenType; var T: PToken): Boolean;
	function GetChar(var C: Char): Boolean;
    function HasData: Boolean;
    function GetStreamInfo: String;
    function GetToken(var T: PToken): Boolean;
    function PeekChar(var C: Char): Boolean;
    function ReadCommentType1(var T: PToken): Boolean;
    function ReadCommentType2(var T: PToken): Boolean;
    function ReadCommentType3(var T: PToken): Boolean;
    function ReadLiteralString(var T: PToken): Boolean;
    function ReadToken(C: Char; S: TCharSet; TT: TTokenType; var T: PToken): Boolean;
  end;

const
  TokenTypeNames: Array[TOK_WHITESPACE..TOK_RESERVED] of String[15] =
    ('whitespace', 'comment', 'identifier', 'number', 'string',
     'symbol', 'directive', 'reserved');
  KEY_AND = 0;
  KEY_ARRAY = 1;
  KEY_AS = 2;
  KEY_ASM = 3;
  KEY_BEGIN = 4;
  KEY_CASE = 5;
  KEY_CLASS = 6;
  KEY_CONST = 7;
  KEY_CONSTRUCTOR = 8;
  KEY_DESTRUCTOR = 9;
  KEY_DIV = 10;
  KEY_DO = 11;
  KEY_DOWNTO = 12;
  KEY_ELSE = 13;
  KEY_END = 14;
  KEY_EXCEPT = 15;
  KEY_EXPORTS = 16;
  KEY_FILE = 17;
  KEY_FINALIZATION = 18;
  KEY_FINALLY = 19;
  KEY_FOR = 20;
  KEY_FUNCTION = 21;
  KEY_GOTO = 22;
  KEY_IF = 23;
  KEY_IMPLEMENTATION = 24;
  KEY_IN = 25;
  KEY_INHERITED = 26;
  KEY_INITIALIZATION = 27;
  KEY_INLINE = 28;
  KEY_INTERFACE = 29;
  KEY_IS = 30;
  KEY_LABEL = 31;
  KEY_LIBRARY = 32;
  KEY_MOD = 33;
  KEY_NIL = 34;
  KEY_NOT = 35;
  KEY_OBJECT = 36;
  KEY_OF = 37;
  KEY_ON = 38;
  KEY_OR = 39;
  KEY_PACKED = 40;
  KEY_PROCEDURE = 41;
  KEY_PROGRAM = 42;
  KEY_PROPERTY = 43;
  KEY_RAISE = 44;
  KEY_RECORD = 45;
  KEY_REPEAT = 46;
  KEY_RESOURCESTRING = 47;
  KEY_SET = 48;
  KEY_SHL = 49;
  KEY_SHR = 50;
  KEY_STRING = 51;
  KEY_THEN = 52;
  KEY_THREADVAR = 53;
  KEY_TO = 54;
  KEY_TRY = 55;
  KEY_TYPE = 56;
  KEY_UNIT = 57;
  KEY_UNTIL = 58;
  KEY_USES = 59;
  KEY_VAR = 60;
  KEY_WHILE = 61;
  KEY_WITH = 62;
  KEY_XOR = 63;
  NUM_KEYS = 64;
  KeyWordArray: Array[0..NUM_KEYS-1] of String[15] =
  ('AND', 'ARRAY', 'AS', 'ASM', 'BEGIN', 'CASE', 'CLASS', 'CONST',
   'CONSTRUCTOR', 'DESTRUCTOR', 'DIV', 'DO', 'DOWNTO', 'ELSE', 'END',
   'EXCEPT', 'EXPORTS', 'FILE', 'FINALIZATION', 'FINALLY', 'FOR', 'FUNCTION',
   'GOTO', 'IF', 'IMPLEMENTATION', 'IN', 'INHERITED', 'INITIALIZATION',
   'INLINE', 'INTERFACE', 'IS', 'LABEL', 'LIBRARY', 'MOD', 'NIL', 'NOT',
   'OBJECT', 'OF', 'ON', 'OR', 'PACKED', 'PROCEDURE', 'PROGRAM', 'PROPERTY',
   'RAISE', 'RECORD', 'REPEAT', 'RESOURCESTRING', 'SET', 'SHL', 'SHR',
   'STRING', 'THEN', 'THREADVAR', 'TO', 'TRY', 'TYPE', 'UNIT', 'UNTIL', 'USES',
   'VAR', 'WHILE', 'WITH', 'XOR');
  KeyWords: PStringCollection = nil;
  SD_ABSOLUTE = 0;
  SD_ABSTRACT = 1;
  SD_APIENTRY = 2;
  SD_ASSEMBLER = 3;
  SD_AUTOMATED = 4;
  SD_CDECL = 5;
  SD_DEFAULT = 6;
  SD_DISPID = 7;
  SD_DYNAMIC = 8;
  SD_EXPORT = 9;
  SD_EXTERNAL = 10;
  SD_FAR = 11;
  SD_FORWARD = 12;
  SD_INDEX = 13;
  SD_MESSAGE = 14;
  SD_NAME = 15;
  SD_NEAR = 16;
  SD_NODEFAULT = 17;
  SD_OVERLOAD = 18;
  SD_OVERRIDE = 19;
  SD_PASCAL = 20;
  SD_PRIVATE = 21;
  SD_PROTECTED = 22;
  SD_PUBLIC = 23;
  SD_PUBLISHED = 24;
  SD_READ = 25;
  SD_REGISTER = 26;
  SD_RESIDENT = 27;
  SD_STDCALL = 28;
  SD_STORED = 29;
  SD_VIRTUAL = 30;
  SD_WRITE = 31;
  NUM_STANDARD_DIRECTIVES = 32;
  StandardDirectiveArray: Array[0..NUM_STANDARD_DIRECTIVES-1] of String[15] =
  ('ABSOLUTE', 'ABSTRACT', 'APIENTRY', 'ASSEMBLER', 'AUTOMATED', 'CDECL',
   'DEFAULT','DISPID', 'DYNAMIC', 'EXPORT', 'EXTERNAL', 'FAR', 'FORWARD',
   'INDEX', 'MESSAGE', 'NAME', 'NEAR', 'NODEFAULT', 'OVERLOAD', 'OVERRIDE',
   'PASCAL', 'PRIVATE', 'PROTECTED', 'PUBLIC', 'PUBLISHED', 'READ', 'REGISTER',
   'RESIDENT', 'STDCALL', 'STORED', 'VIRTUAL', 'WRITE');
  StandardDirectives: PStringCollection = nil;

implementation

uses
  arrays, numbers;

const
  Whitespace = [#9, #10, #13, ' '];
  Letters = ['A'..'Z', 'a'..'z'];
  DecimalDigits = ['0'..'9'];
  HexadecimalDigits = DecimalDigits + ['A'..'F', 'a'..'f'];
  IdentifierStart = ['_'] + Letters;
  IdentifierOther = IdentifierStart + DecimalDigits;
  CharOther = HexadecimalDigits + ['$'];
  NumberStart = DecimalDigits + ['$'];
  NumberOther = HexadecimalDigits + ['.', '+', '-'];
  QuoteChar = '''';
  NUM_SINGLE_CHAR_SYMBOLS = 10;
  SingleCharSymbols: array[0..NUM_SINGLE_CHAR_SYMBOLS - 1] of
    record
      C: Char;
      S: TSymbolType;
	end =
    ((C: ';'; S: SYM_SEMICOLON),
     (C: ','; S: SYM_COMMA),
     (C: '['; S: SYM_LEFT_BRACKET),
     (C: ']'; S: SYM_RIGHT_BRACKET),
     (C: '+'; S: SYM_PLUS),
     (C: '-'; S: SYM_MINUS),
     (C: '*'; S: SYM_ASTERISK),
     (C: '='; S: SYM_EQUAL),
     (C: '^'; S: SYM_ROOF),
     (C: '@'; S: SYM_AT));

{ TToken }

constructor TToken.Init(TT: TTokenType);
begin
  inherited Init;
  MyType := TT;
end;

function TToken.GetTypeName: String;
begin
  GetTypeName := TOKEN_TYPE_NAMES[MyType];
end;

function TToken.IsSymbol(ST: TSymbolType): Boolean;
begin
  IsSymbol := (MyType = TOK_SYMBOL) and (Info.SymbolType = ST);
end;

{ TTokenizer }

constructor TTokenizer.Init(S: PInputStream);
begin
  inherited Init;
  EOS := false;
  ErrorMessage := '';
  IsCharBuffered := false;
  Row := 1;
  Stream := S;
end;

destructor TTokenizer.Done;
begin
  if Assigned(Stream)
  then Dispose(Stream, Done);
  inherited Done;
end;

procedure TTokenizer.CheckForDirective(var T: TToken; Offset: Integer);
begin
  if Assigned(T.Data) and (T.Data^.Content > Offset) and
    (T.Data^.Data^[Offset] = '$')
  then
    begin
      T.MyType := TOK_DIRECTIVE;
      {WriteLn('Found directive: ', T.Data^.GetString);}
    end;
end;

procedure TTokenizer.ConsumeChar;
begin
  IsCharBuffered := false;
end;

function TTokenizer.CreateSymbolToken(ST: TSymbolType; S: String; var T: PToken): Boolean;
begin
  if (not CreateToken(TOK_SYMBOL, T))
  then
    begin
      CreateSymbolToken := false;
      exit;
    end;
  T^.Data^.AppendString(S);
  T^.Info.SymbolType := ST;
  CreateSymbolToken := true;
end;

function TTokenizer.CreateToken(TT: TTokenType; var T: PToken): Boolean;
begin
  CreateToken := false;
  T := New(PToken, Init(TT));
  if (not Assigned(T))
  then exit;
  T^.Data := New(PText, Init);
  if (not Assigned(T^.Data))
  then
    begin
      Dispose(T, Done);
      T := nil;
      exit;
    end;
  CreateToken := true;
end;

function TTokenizer.GetChar(var C: Char): Boolean;
var
  S: TSInt32;
begin
  if IsCharBuffered
  then
    begin
      C := BufferedChar;
      IsCharBuffered := false;
      GetChar := true;
    end
  else
    begin
      S := Stream^.ReadByte;
      C := Char(S);
      GetChar := (S <> -1);
    end;
end;

function TTokenizer.GetStreamInfo: String;
begin
  GetStreamInfo := Stream^.GetName + '(' + IntToString(Row) + ')';
end;

function TTokenizer.HasData: Boolean;
begin
  HasData := IsCharBuffered or Stream^.HasData;
end;

function TTokenizer.GetToken(var T: PToken): Boolean;
var
  C: Char;
  I: Integer;
  S: String{[255]};
begin
  GetToken := false;
  T := nil;
  if (not GetChar(C))
  then exit;
  if (C in Whitespace)
  then
    begin
	  if ReadToken(C, Whitespace, TOK_WHITESPACE, T)
      then
        begin
          { after successful reading all whitespace characters, update
            internal row counter to be able to state current row on errors;
            caveat: will fail on Mac files (row is 13) }
          I := 0;
          while (I < T^.Data^.Content) do
          begin
            if (T^.Data^.Data^[I] = #10)
            then Inc(Row);
            Inc(I);
          end;
          GetToken := true;
        end
      else GetToken := false;
    end
  else
  if (C in IdentifierStart)
  then
    begin
	  if ReadToken(C, IdentifierOther, TOK_IDENTIFIER, T)
      then
        begin
          { check if identifier is a reserved identifier }
          S := T^.Data^.GetString;
          Arrays.StringToUpper(S, S);
          I := KeyWords^.IndexOf(@S);
          if (I <> -1)
          then
            begin
              T^.MyType := TOK_RESERVED;
              T^.Info.ReservedKey := I;
            end;
          GetToken := true;
        end
    end
  else
  if (C in NumberStart)
  then GetToken := ReadToken(C, NumberOther, TOK_NUMBER, T)
  else
    case C of
      QuoteChar: GetToken := ReadLiteralString(T);
      '#': GetToken := ReadToken(C, CharOther, TOK_STRING, T);
      '{':
	  begin
	    if ReadCommentType1(T)
        then
		  begin
		    CheckForDirective(T^, 1);
            GetToken := true;
          end;
      end;
      '(':
      begin
        C := ' ';
        if HasData and (not PeekChar(C))
        then exit;
        case C of
          '*':
          begin
            ConsumeChar;
            if ReadCommentType2(T)
            then
              begin
                CheckForDirective(T^, 2);
                GetToken := true;
              end;
          end;
          '.':
          begin
            ConsumeChar;
            GetToken := CreateSymbolToken(SYM_LEFT_BRACKET, '(.', T);
          end;
          else GetToken := CreateSymbolToken(SYM_LEFT_PARENTHESIS, '(', T);
        end;
      end;
      ')':
      begin
        C := ' ';
        GetToken := CreateSymbolToken(SYM_RIGHT_PARENTHESIS, ')', T);
      end;
      '.':
      begin
        C := ' ';
        if HasData and (not PeekChar(C))
        then exit;
        case C of
          '.':
          begin
            ConsumeChar;
            GetToken := CreateSymbolToken(SYM_RANGE, '..', T);
		  end;
		  ')':
		  begin
            ConsumeChar;
            GetToken := CreateSymbolToken(SYM_RIGHT_BRACKET, '.)', T);
		  end;
          else GetToken := CreateSymbolToken(SYM_PERIOD, '.', T);
        end;
      end;
      '/':
      begin
        C := ' ';
        if HasData and (not PeekChar(C))
        then exit;
        case C of
          '/':
          begin
            ConsumeChar;
            GetToken := ReadCommentType3(T);
          end;
          else GetToken := CreateSymbolToken(SYM_SLASH, '/', T);
        end;
      end;
      ':':
      begin
        C := ' ';
        if HasData and (not PeekChar(C))
        then exit;
        case C of
          '=':
          begin
            ConsumeChar;
            GetToken := CreateSymbolToken(SYM_ASSIGN, ':=', T);
		  end;
          else GetToken := CreateSymbolToken(SYM_COLON, ':', T);
        end;
      end;
      '<':
      begin
        C := ' ';
        if HasData and (not PeekChar(C))
        then exit;
        case C of
          '=':
          begin
            ConsumeChar;
            GetToken := CreateSymbolToken(SYM_LESS_THAN_EQUAL, '<=', T);
          end;
          else GetToken := CreateSymbolToken(SYM_LESS_THAN, '<', T);
        end;
      end;
      '>':
      begin
        C := ' ';
        if HasData and (not PeekChar(C))
        then exit;
        case C of
          '=':
          begin
            ConsumeChar;
            GetToken := CreateSymbolToken(SYM_GREATER_THAN_EQUAL, '<=', T);
          end;
          else GetToken := CreateSymbolToken(SYM_GREATER_THAN, '<', T);
        end;
      end;
      else
        begin
          for I := 0 to NUM_SINGLE_CHAR_SYMBOLS - 1 do
          begin
            if (C = SingleCharSymbols[I].C)
            then
              begin
                GetToken := CreateSymbolToken(SingleCharSymbols[I].S,
				  SingleCharSymbols[I].C, T);
                exit;
              end;
          end;
          ErrorMessage := 'Error: Invalid character in Pascal input stream - ';
        end;
    end;
end;

function TTokenizer.PeekChar(var C: Char): Boolean;
var
  S: TSInt32;
begin
  if IsCharBuffered
  then
    begin
      C := BufferedChar;
      PeekChar := true;
    end
  else
    begin
      if (Stream^.HasData)
      then
        begin
          S := Stream^.ReadByte;
          C := Char(S);
          BufferedChar := C;
          IsCharBuffered := true;
          PeekChar := true;
        end
      else
        begin
          EOS := true;
          PeekChar := false;
        end;
    end;
end;

function TTokenizer.ReadCommentType1(var T: PToken): Boolean;
var
  C: Char;
begin
  ReadCommentType1 := false;
  if (not CreateToken(TOK_COMMENT, T))
  then exit;
  T^.Data^.AppendChar('{');
  repeat
    if (not HasData) or (not GetChar(C))
    then
      begin
        exit;
      end;
    if (C = #10)
    then Inc(Row);
    if (C = '{')
    then WriteLn(GetStreamInfo + 'Warning: Found opening brace { within comment, nested comments are not supported.');
    T^.Data^.AppendChar(C);
  until (C = '}');
  ReadCommentType1 := true;
end;

function TTokenizer.ReadCommentType2(var T: PToken): Boolean;
var
  C: Char;
  Finished: Boolean;
begin
  ReadCommentType2 := false;
  if (not CreateToken(TOK_COMMENT, T))
  then exit;
  T^.Data^.AppendString('(*');
  Finished := false;
  repeat
    if (not HasData) or (not GetChar(C))
    then
      begin
        exit;
      end;
    T^.Data^.AppendChar(C);
    if (C = #10)
    then Inc(Row);
    if (C = '*')
    then
      begin
        if (not HasData) or (not PeekChar(C))
        then
          begin
            exit;
          end;
        if (C = ')')
        then
          begin
            ConsumeChar;
            T^.Data^.AppendChar(C);
            Finished := true;
          end;
      end;
  until Finished;
  ReadCommentType2 := true;
end;

function TTokenizer.ReadCommentType3(var T: PToken): Boolean;
var
  C: Char;
  Finished: Boolean;
begin
  ReadCommentType3 := false;
  if (not CreateToken(TOK_COMMENT, T))
  then exit;
  T^.Data^.AppendString('//');
  Finished := false;
  repeat
    if (not HasData) or (not GetChar(C))
    then
      begin
        exit;
      end;
    if (C <> #10) and (C <> #13)
    then T^.Data^.AppendChar(C)
    else Finished := true;
  until Finished;
  ReadCommentType3 := true;
end;

function TTokenizer.ReadLiteralString(var T: PToken): Boolean;

  procedure ReleaseToken;
  begin
    Dispose(T, Done);
    T := nil;
  end;

var
  C: Char;
  Finished: Boolean;
begin
  ReadLiteralString := false;
  Finished := false;
  if (not CreateToken(TOK_STRING, T))
  then exit;
  T^.Data := New(PText, Init);
  if (not Assigned(T^.Data))
  then
    begin
      ReleaseToken;
      exit;
    end;
  T^.Data^.AppendChar('''');
  repeat
    if (not Stream^.HasData)
    then
      begin
        ErrorMessage := 'Error: tokenizer: unexpected end of stream.';
        ReleaseToken;
        exit;
      end;
    if (not GetChar(C))
    then
      begin
        ErrorMessage := 'Error: tokenizer: could not read character.';
        ReleaseToken;
        exit;
      end;
    if (C = QuoteChar)
    then
      begin
        if (not PeekChar(C))
        then
          begin
            ErrorMessage := 'Error: tokenizer: could not peek character.';
            ReleaseToken;
            exit;
          end;
        if (C = QuoteChar)
        then { escaped single quote within string }
          begin
            ConsumeChar;
            T^.Data^.AppendChar(QuoteChar);
          end
        else { end of string }
          begin
            Finished := true;
          end;
        T^.Data^.AppendChar(QuoteChar);
      end
    else
      begin
        T^.Data^.AppendChar(C);
      end;
  until Finished;
  ReadLiteralString := true;
end;

function TTokenizer.ReadToken(C: Char; S: TCharSet; TT: TTokenType; var T: PToken): Boolean;
begin
  ReadToken := false;
  if (not CreateToken(TT, T))
  then exit;
  T^.Data^.AppendChar(C);
  repeat
    if (not PeekChar(C))
    then
      begin
        if EOS
        then
          begin
            ReadToken := true;
          end
        else
          begin
            Dispose(T, Done);
            T := nil;
          end;
        exit;
      end;
    if (C in S)
    then
      begin
        T^.Data^.AppendChar(C);
        ConsumeChar;
      end
    else
      begin
        ReadToken := true;
        exit;
      end;
  until false;
end;

{ global functions and procedures }

procedure InitArrays;
var
  I: Integer;
begin
  KeyWords := New(PStringCollection, Init(NUM_KEYS, 1));
  for I := 0 to NUM_KEYS - 1 do
  begin
    KeyWords^.Insert(NewStr(KeyWordArray[I]));
  end;
  StandardDirectives := New(PStringCollection, Init(NUM_STANDARD_DIRECTIVES, 1));
  for I := 0 to NUM_STANDARD_DIRECTIVES - 1 do
  begin
    StandardDirectives^.Insert(NewStr(StandardDirectiveArray[I]));
  end;
end;

begin
  InitArrays;
end.
